from bs4 import BeautifulSoup
import os
import config
import re
import spacy
import numpy as np

from model import Mention, Entity

nlp = spacy.load("en_core_web_lg")

def from_isnotes(path):
    
    split_path = path.split(r"/")
    root, file_name = "/".join(split_path[:-1]), split_path[-1]
    
    mentions = []
    
    if file_name.endswith(".mmax"):
        
        #print(file_name)
        
        file_root = file_name[:-5]
        
        basedata = BeautifulSoup(open(root+"/Basedata/"+file_root+"_words.xml", "r").read(), features="lxml")
        entities = BeautifulSoup(open(root+"/markables/"+file_root+"_entity_level.xml", "r").read(), features="lxml")
        coref = BeautifulSoup(open(root+"/markables/"+file_root+"_coref_level.xml", "r").read(), features="lxml")
        
        for entity in entities.contents[2].contents[0].contents[0].find_all():
            
            span = entity["span"]
            tokens = []
            mention_id = entity["id"]
            
            # comparative: either "outsidetext" or antecedent id
            comp_from = None
            if entity["information_status"] == "mediated":
                try:
                    if entity["mediated_type"] == "comparative":
                        comp_from = entity["comparative_type"]
                        if comp_from == "withintext":
                            comp_from = entity["comp_from"]
                except:
                    pass
            
            try:
                coref_set = coref.find("markable", {"span": span})["coref_set"]
            except:
                coref_set = None
            
            if ".." in span:
                
                span_ids = [int(re.sub("word_", "", word)) for word in span.split("..")]
                for i in range(span_ids[0],span_ids[1]+1):
                    token = basedata.find("word", {"id": "word_"+str(i)}).contents[0]
                    tokens.append(token)
                
                mentions.append(Mention( tokens, [span_ids[0],span_ids[1]], mention_id, coref_set, comp_from))
                
            else:
                
                token = basedata.find("word", {"id": span}).contents[0]
                tokens.append(token)
                
                word_id = int(span.split("_")[-1])
                mentions.append(Mention(tokens, [word_id, word_id], mention_id, coref_set, comp_from))
    
    return mentions

def from_conll(path):
    
    conll = open(path, 'r') # source in tsv format
    mentions = [] # return list of Mention objects
    word_id = 0
    
    for line in conll:
        if line.startswith("#begin document"):
            sentence_id = 0
            coref_spans = dict() # coref spans are not cross document
            text = [] # new document also means new sentence
            tags = [] # and new pos tags
            coref_stack = {}
            
            mention_id = 0
            
        elif line.startswith("#end document"):
            pass
        
        elif line == "\n":
            sentence_id = 0
            #doc = nlp(" ".join(text)) 
            doc = nlp.tokenizer.tokens_from_list(text)
            nlp.tagger(doc)
            nlp.parser(doc)
            
            for chunk in doc.noun_chunks:
                tokens = chunk.text.split(" ")
                left_edge = word_id + chunk.root.left_edge.i
                right_edge = word_id + chunk.root.right_edge.i
                
                if (left_edge, right_edge) in coref_spans:
                    coref_set = coref_spans[(left_edge, right_edge)]
                else:
                    coref_set = None
                
                if chunk.root.head.text.lower() == "others":
                        comp_from = True
                else: 
                        comp_from = False
                        
                for token in chunk:
                    
                    if token.dep_ == "poss" and token.head == chunk.root:
                        left_poss = word_id + token.left_edge.i
                        right_poss = word_id + token.right_edge.i
                
                        if (left_poss, right_poss) in coref_spans:
                            poss_coref = coref_spans[(left_poss, right_poss)]
                        else:
                            poss_coref = None
                        
                        new_mention = Mention([child.text for child in token.children],
                                              [left_poss, right_poss], mention_id, poss_coref, False)
                        mentions.append(new_mention)
                        mention_id+=1
                        
                    if token.text.lower() == "than":
                        comp_from = False
                        break
                    elif tags[token.i] == "JJR":
                        comp_from = True
                    elif token.text.lower() == "more" and token.head.pos_ == "ADJ":
                        comp_from = True
                    elif token.text.lower() == "less" and token.head.pos_ == "ADJ":
                        comp_from = True
                    elif token.text.lower() in config.COMPARATIVES and token.head == chunk.root:
                        comp_from = True
                    
                new_mention = Mention(tokens, [left_edge, right_edge], 
                                      mention_id, coref_set, comp_from)
                mentions.append(new_mention)
                mention_id += 1
                
            word_id += len(text)
            text = [] # linebreak means new sentence
            tags = []
            coref_stack = {}
            
        else:
            naked_line = line.strip()
            split_line = naked_line.split()
            # Important indices:
            # 1:doc_id, 2:token_id,     3:token,
            # 4:POS,    5:syntax_tree, -1:Coref
        
            coref = split_line[-1].split("|")
            
            coref_stack = {**coref_stack, **{int(c): word_id+sentence_id for c in re.findall("\(([0-9]+)", split_line[-1])}}
            
            for c in re.findall("([0-9]+)\)", split_line[-1]):
                if int(c) not in coref_stack:
                    raise Exception
                coref_spans[(coref_stack[int(c)], word_id+sentence_id)] = int(c)
            
            text.append(split_line[3])
            tags.append(split_line[4])
            
            sentence_id += 1
            
    return mentions
            
            
def to_mmax(source_path, target_directory_path):
    """ Source must be in conll format
    
    """
    
    name = os.path.splitext(os.path.basename(source_path))
    base_name = name[0]
    
    print("\n[NCR] Extracting entities from \033[1m{}\033[0m.".format("".join(name)))
    
    mentions = from_conll(source_path)
    
    comp_from_list = [m.comp_from for m in mentions]
    
    len_mentions = sum(bool(c) for c in comp_from_list)
        
    
    new_name = "{:0>2}_{}".format(len_mentions, base_name)
    
    if any(comp_from_list):
        
        mmax = open(os.path.join(target_directory_path,  new_name+".mmax"), "w")
        mmax.write("""<?xml version="1.0"?>
<mmax_project>
<!--<sentences>002_htc_text.xml</sentences>-->
<turns></turns>
<words>{}</words>
<gestures></gestures>
<keyactions></keyactions>
</mmax_project>
""".format(new_name+"_words.xml"))
        mmax.close()
        
        target_path = os.path.join(target_directory_path, "Basedata", new_name+"_words.xml")
        
        source = open(source_path, 'r')
        target = open(target_path, 'w')
        
        word_id = 0
        
        
        target.write(r"""<?xml version='1.0' encoding='UTF-8'?>
<!DOCTYPE words SYSTEM "words.dtd">
<words>""")
        
        for line in source:
            if line.startswith("#begin document"):
                pass
            elif line.startswith("#end document"):
                pass
            elif line == "\n":
                pass
            else:
                naked_line = line.strip()
                split_line = naked_line.split()
                
                token = split_line[3]
                if token == "&":
                    token = r"&amp;"
                token = re.sub("<", "&lt;", token)
                token = re.sub(">", "&gt;", token)
                
                target.write(r'<word id="word_'+str(word_id)+r'">'+token+r'</word>'+'\n')
                
                word_id += 1
        
        target.write(r"</words>")
        
        print("[NCR] Writing \033[92m{:>4}\033[0m token(s) to \033[1m{}_words.xml\033[0m.".format(word_id, new_name))
        
        markable_id = 0

        entity_level_path = os.path.join(target_directory_path+"Markables", new_name+"_entity_level.xml")
        entity_level = open(entity_level_path, 'w')
        entity_level.write(r"""<?xml version='1.0'?>
<!DOCTYPE markables SYSTEM "markables.dtd">
<markables xmlns="www.comp.leeds.ac.uk/markert/entity">""")
        
        coref_level_path = os.path.join(target_directory_path+"Markables", new_name+"_coref_level.xml")
        coref_level = open(coref_level_path, 'w')
        coref_level.write(r"""<?xml version='1.0'?>
<!DOCTYPE markables SYSTEM "markables.dtd">
<markables xmlns="www.eml.org/NameSpaces/coref">""")
        
        coref_chains = set()
        
        for mention in mentions:
            if mention.comp_from:
                entity_level.write(r'<markable id="markable_'+str(markable_id)+r'" information_status="mediated" mediated_type="comparative" comparative_type="withintext" span="word_'+str(mention.span[0])+r'..word_'+str(mention.span[1])+r'" mmax_level="entity" />'+'\n')
                markable_id += 1
            else:
                entity_level.write(r'<markable id="markable_'+str(markable_id)+r'" span="word_'+str(mention.span[0])+r'..word_'+str(mention.span[1])+r'" mmax_level="entity" />'+'\n')
                markable_id += 1
            
            if mention.coref_set:
                coref_chains.add(mention.coref_set)
                coref_level.write(r'<markable id="markable_'+str(markable_id)+r'" span="word_'+str(mention.span[0])+r'..word_'+str(mention.span[1])+r'" coref_set="set_'+str(mention.coref_set)+r'" mmax_level="coref" />'+'\n')
                markable_id += 1
            else:
                pass
        
        entity_level.write(r'</markables>')
        coref_level.write(r'</markables>')
        
        entity_level.close()
        coref_level.close()
    
        print("[NCR] Writing \033[92m{:>4}\033[0m mention(s) to \033[1m{}_entity_level.xml\033[0m.".format(len_mentions, new_name))
        print("[NCR] Writing \033[92m{:>4}\033[0m coreference chain(s) to \033[1m{}_coref_level.xml\033[0m.".format(len(coref_chains), new_name))
        
    else:
        print("[NCR] \033[1m{}\033[96m does not contain any comparatives.\033[0m".format(base_name))
        
    
if __name__ == "__main__":
    # ISnotes
    #root = config.DATA_PATH
    #files = os.listdir(root)
    
    #for f in files:
        #get_mentions(root+f)
    
    # conll/OntoNotes
    root = config.CONLL_PATH
    
    files = []
    for x in os.walk(root):
        for y in x[2]:
            #print(y)
            if y.endswith(".v4_gold_conll"):
                files.append(x[0]+"/"+y)
    
    for f in files:
        to_mmax(f, "/home/students/zimmermann/Projects/ncr/annotation/Entity/")
    
    
